import logging

from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import backref, relationship

from fonduer.meta import Meta
from fonduer.utils.utils import camel_to_under

_meta = Meta.init()
logger = logging.getLogger(__name__)

# This global dictionary contains all classes that have been declared in this
# Python environment, so that mention_subclass() can return a class if it
# already exists and is identical in specification to the requested class.
mention_subclasses = {}


class Mention(_meta.Base):
    """
    An abstract Mention.

    New mention types should be defined by calling mention_subclass(),
    **not** subclassing this class directly.
    """

    __tablename__ = "mention"

    #: The unique id of the ``Mention``.
    id = Column(Integer, primary_key=True)

    #: The type for the ``Mention``, which corresponds to the names the user
    #: gives to the mention_subclass.
    type = Column(String, nullable=False)

    __mapper_args__ = {"polymorphic_identity": "mention", "polymorphic_on": type}

    def get_contexts(self):
        """Get the constituent context making up this mention."""
        return tuple(getattr(self, name) for name in self.__argnames__)

    def __len__(self):
        return len(self.__argnames__)

    def __getitem__(self, key):
        return self.get_contexts()[key]

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(map(str, self.get_contexts())),
        )

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()


def mention_subclass(class_name, cardinality=None, values=None, table_name=None):
    """
    Creates and returns a Mention subclass with provided argument names,
    which are Context type. Creates the table in DB if does not exist yet.

    Import using:

    .. code-block:: python

        from fonduer.candidates.models import mention_subclass

    :param class_name: The name of the class, should be "camel case" e.g.
        NewMention
    :param table_name: The name of the corresponding table in DB; if not
        provided, is converted from camel case by default, e.g. new_mention
    :param values: The values that the variable corresponding to the Mention
        can take. By default it will be [True, False].
    :param cardinality: The cardinality of the variable corresponding to the
        Mention. By default is 2 i.e. is a binary value, e.g. is or is not
        a true mention.
    """
    if table_name is None:
        table_name = camel_to_under(class_name)

    # If cardinality and values are None, default to binary classification
    if cardinality is None and values is None:
        values = [True, False]
        cardinality = 2
    # Else use values if present, and validate proper input
    elif values is not None:
        if cardinality is not None and len(values) != cardinality:
            raise ValueError("Number of values must match cardinality.")
        if None in values:
            raise ValueError("`None` is a protected value.")
        # Note that bools are instances of ints in Python...
        if any([isinstance(v, int) and not isinstance(v, bool) for v in values]):
            raise ValueError(
                (
                    "Default usage of values is consecutive integers."
                    "Leave values unset if trying to define values as integers."
                )
            )
        cardinality = len(values)

    # If cardinality is specified but not values, fill in with ints
    elif cardinality is not None:
        values = list(range(cardinality))

    args = ["context"]
    class_spec = (args, table_name, cardinality, values)
    if class_name in mention_subclasses:
        if class_spec == mention_subclasses[class_name][1]:
            return mention_subclasses[class_name][0]
        else:
            raise ValueError(
                "Mention subclass "
                + class_name
                + " already exists in memory with incompatible "
                + "specification: "
                + str(mention_subclasses[class_name][1])
            )
    else:
        # Set the class attributes == the columns in the database
        class_attribs = {
            # Declares name for storage table
            "__tablename__": table_name,
            # Connects mention_subclass records to generic Mention records
            "id": Column(
                Integer, ForeignKey("mention.id", ondelete="CASCADE"), primary_key=True
            ),
            # Store values & cardinality information in the class only
            "values": values,
            "cardinality": cardinality,
            # Polymorphism information for SQLAlchemy
            "__mapper_args__": {"polymorphic_identity": table_name},
            # Helper method to get argument names
            "__argnames__": args,
        }
        class_attribs["document_id"] = Column(
            Integer, ForeignKey("document.id", ondelete="CASCADE")
        )
        class_attribs["document"] = relationship(
            "Document",
            backref=backref(table_name + "s", cascade="all, delete-orphan"),
            foreign_keys=class_attribs["document_id"],
        )

        # Create named arguments, i.e. the entity mentions comprising the
        # relation mention.
        unique_args = []
        for arg in args:

            # Primary arguments are constituent Contexts, and their ids
            class_attribs[arg + "_id"] = Column(
                Integer, ForeignKey("context.id", ondelete="CASCADE")
            )
            class_attribs[arg] = relationship(
                "Context", foreign_keys=class_attribs[arg + "_id"]
            )
            unique_args.append(class_attribs[arg + "_id"])

        # Add unique constraints to the arguments
        class_attribs["__table_args__"] = (UniqueConstraint(*unique_args),)

        # Create class
        C = type(class_name, (Mention,), class_attribs)

        # Create table in DB
        if not _meta.engine.dialect.has_table(_meta.engine, table_name):
            C.__table__.create(bind=_meta.engine)

        mention_subclasses[class_name] = C, class_spec

        return C



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryCellMention(TemporaryContext):
    """The TemporaryContext version of CellMention."""

    def __init__(self, cell):
        super(TemporaryCellMention, self).__init__()
        self.cell = cell  # The cell Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.cell == other.cell
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.cell != other.cell
        except AttributeError:
            return True

    def __contains__(self, other_cell):
        return self.__eq__(other_cell)

    def __hash__(self):
        return hash(self.cell)

    def get_stable_id(self):
        """Return a stable id for the ``CellMention``."""
        return "%s::%s:%s:%s" % (
            self.cell.document.name,
            self._get_polymorphic_identity(),
            self.cell.table.position,
            self.cell.position,
        )

    def _get_table(self):
        return CellMention

    def _get_polymorphic_identity(self):
        return "cell_mention"

    def _get_insert_args(self):
        return {"cell_id": self.cell.id}

    def __repr__(self):
        return "{}(document={}, table_position={}, position={})".format(
            self.__class__.__name__,
            self.cell.document.name,
            self.cell.table.position,
            self.cell.position,
        )

    def _get_instance(self, **kwargs):
        return TemporaryCellMention(**kwargs)


class CellMention(Context, TemporaryCellMention):
    """A cell ``Mention``."""

    __tablename__ = "cell_mention"

    #: The unique id of the ``CellMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Cell``.
    cell_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Cell``.
    cell = relationship("Context", foreign_keys=cell_id)

    __table_args__ = (UniqueConstraint(cell_id),)

    __mapper_args__ = {
        "polymorphic_identity": "cell_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship
from sqlalchemy.sql import text
from sqlalchemy.types import PickleType

from fonduer.candidates.models.span_mention import TemporarySpanMention
from fonduer.parser.models.context import Context
from fonduer.parser.models.utils import split_stable_id


class TemporaryImplicitSpanMention(TemporarySpanMention):
    """The TemporaryContext version of ImplicitSpanMention."""

    def __init__(
        self,
        sentence,
        char_start,
        char_end,
        expander_key,
        position,
        text,
        words,
        lemmas,
        pos_tags,
        ner_tags,
        dep_parents,
        dep_labels,
        page,
        top,
        left,
        bottom,
        right,
        meta=None,
    ):
        super(TemporarySpanMention, self).__init__()
        self.sentence = sentence  # The sentence Context of the Span
        self.char_start = char_start
        self.char_end = char_end
        self.expander_key = expander_key
        self.position = position
        self.text = text
        self.words = words
        self.lemmas = lemmas
        self.pos_tags = pos_tags
        self.ner_tags = ner_tags
        self.dep_parents = dep_parents
        self.dep_labels = dep_labels
        self.page = page
        self.top = top
        self.left = left
        self.bottom = bottom
        self.right = right
        self.meta = meta

    def __len__(self):
        return sum(map(len, self.words))

    def __eq__(self, other):
        try:
            return (
                self.sentence == other.sentence
                and self.char_start == other.char_start
                and self.char_end == other.char_end
                and self.expander_key == other.expander_key
                and self.position == other.position
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return (
                self.sentence != other.sentence
                or self.char_start != other.char_start
                or self.char_end != other.char_end
                or self.expander_key != other.expander_key
                or self.position != other.position
            )
        except AttributeError:
            return True

    def __hash__(self):
        return (
            hash(self.sentence)
            + hash(self.char_start)
            + hash(self.char_end)
            + hash(self.expander_key)
            + hash(self.position)
        )

    def get_stable_id(self):
        """
        Return a stable id.

        :rtype: string
        """
        doc_id, _, parent_doc_char_start, _ = split_stable_id(self.sentence.stable_id)
        return "%s::%s:%s:%s:%s:%s" % (
            self.sentence.document.name,
            self._get_polymorphic_identity(),
            parent_doc_char_start + self.char_start,
            parent_doc_char_start + self.char_end,
            self.expander_key,
            self.position,
        )

    def _get_table(self):
        return ImplicitSpanMention

    def _get_polymorphic_identity(self):
        return "implicit_span_mention"

    def _get_insert_args(self):
        return {
            "sentence_id": self.sentence.id,
            "char_start": self.char_start,
            "char_end": self.char_end,
            "expander_key": self.expander_key,
            "position": self.position,
            "text": self.text,
            "words": self.words,
            "lemmas": self.lemmas,
            "pos_tags": self.pos_tags,
            "ner_tags": self.ner_tags,
            "dep_parents": self.dep_parents,
            "dep_labels": self.dep_labels,
            "page": self.page,
            "top": self.top,
            "left": self.left,
            "bottom": self.bottom,
            "right": self.right,
            "meta": self.meta,
        }

    def get_attrib_tokens(self, a="words"):
        """Get the tokens of sentence attribute *a*.

        Intuitively, like calling::

            implicit_span.a


        :param a: The attribute to get tokens for.
        :type a: str
        :return: The tokens of sentence attribute defined by *a* for the span.
        :rtype: list
        """
        return self.__getattribute__(a)

    def get_attrib_span(self, a, sep=" "):
        """Get the span of sentence attribute *a*.

        Intuitively, like calling::

            sep.join(implicit_span.a)

        :param a: The attribute to get a span for.
        :type a: str
        :param sep: The separator to use for the join.
        :type sep: str
        :return: The joined tokens, or text if a="words".
        :rtype: str
        """
        if a == "words":
            return self.text
        else:
            return sep.join(self.get_attrib_tokens(a))

    def __getitem__(self, key):
        """Slice operation returns a new candidate sliced according to **char index**.

        Note that the slicing is w.r.t. the candidate range (not the abs.
        sentence char indexing)
        """
        if isinstance(key, slice):
            char_start = (
                self.char_start if key.start is None else self.char_start + key.start
            )
            if key.stop is None:
                char_end = self.char_end
            elif key.stop >= 0:
                char_end = self.char_start + key.stop - 1
            else:
                char_end = self.char_end + key.stop
            return self._get_instance(
                sentence=self.sentence,
                char_start=char_start,
                char_end=char_end,
                expander_key=self.expander_key,
                position=self.position,
                text=text,
                words=self.words,
                lemmas=self.lemmas,
                pos_tags=self.pos_tags,
                ner_tags=self.ner_tags,
                dep_parents=self.dep_parents,
                dep_labels=self.dep_labels,
                page=self.page,
                top=self.top,
                left=self.left,
                bottom=self.bottom,
                right=self.right,
                meta=self.meta,
            )
        else:
            raise NotImplementedError()

    def __repr__(self):
        return '{}("{}", sentence={}, words=[{},{}], position=[{}])'.format(
            self.__class__.__name__,
            self.get_span(),
            self.sentence.id,
            self.get_word_start_index(),
            self.get_word_end_index(),
            self.position,
        )

    def _get_instance(self, **kwargs):
        return TemporaryImplicitSpanMention(**kwargs)


class ImplicitSpanMention(Context, TemporaryImplicitSpanMention):
    """A span of characters that may not appear verbatim in the source text.

    It is identified by Context id, character-index start and end (inclusive),
    as well as a key representing what 'expander' function drew the ImplicitSpanMention
    from an existing SpanMention, and a position (where position=0 corresponds to the
    first ImplicitSpanMention produced from the expander function).

    The character-index start and end point to the segment of text that was
    expanded to produce the ImplicitSpanMention.
    """

    __tablename__ = "implicit_span_mention"

    #: The unique id of the ``ImplicitSpanMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Sentence``.
    sentence_id = Column(
        Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True
    )
    #: The parent ``Sentence``.
    sentence = relationship("Context", foreign_keys=sentence_id)

    #: The starting character-index of the ``ImplicitSpanMention``.
    char_start = Column(Integer, nullable=False)

    #: The ending character-index of the ``ImplicitSpanMention`` (inclusive).
    char_end = Column(Integer, nullable=False)

    #: The key representing the expander function which produced this
    # ``ImplicitSpanMention``.
    expander_key = Column(String, nullable=False)

    #: The position of the ``ImplicitSpanMention`` where position=0 is the first
    #: ``ImplicitSpanMention`` produced by the expander.
    position = Column(Integer, nullable=False)

    #: The raw text of the ``ImplicitSpanMention``.
    text = Column(String)

    #: A list of the words in the ``ImplicitSpanMention``.
    words = Column(postgresql.ARRAY(String), nullable=False)

    #: A list of the lemmas for each word in the ``ImplicitSpanMention``.
    lemmas = Column(postgresql.ARRAY(String))

    #: A list of the POS tags for each word in the ``ImplicitSpanMention``.
    pos_tags = Column(postgresql.ARRAY(String))

    #: A list of the NER tags for each word in the ``ImplicitSpanMention``.
    ner_tags = Column(postgresql.ARRAY(String))

    #: A list of the dependency parents for each word in the ``ImplicitSpanMention``.
    dep_parents = Column(postgresql.ARRAY(Integer))

    #: A list of the dependency labels for each word in the ``ImplicitSpanMention``.
    dep_labels = Column(postgresql.ARRAY(String))

    #: A list of the page number each word in the ``ImplicitSpanMention``.
    page = Column(postgresql.ARRAY(Integer))

    #: A list of each word's TOP bounding box coordinate in the
    # ``ImplicitSpanMention``.
    top = Column(postgresql.ARRAY(Integer))

    #: A list of each word's LEFT bounding box coordinate in the
    # ``ImplicitSpanMention``.
    left = Column(postgresql.ARRAY(Integer))

    #: A list of each word's BOTTOM bounding box coordinate in the
    # ``ImplicitSpanMention``.
    bottom = Column(postgresql.ARRAY(Integer))

    #: A list of each word's RIGHT bounding box coordinate in the
    # ``ImplicitSpanMention``.
    right = Column(postgresql.ARRAY(Integer))

    #: Pickled metadata about the ``ImplicitSpanMention``.
    meta = Column(PickleType)

    __table_args__ = (
        UniqueConstraint(sentence_id, char_start, char_end, expander_key, position),
    )

    __mapper_args__ = {
        "polymorphic_identity": "implicit_span_mention",
        "inherit_condition": (id == Context.id),
    }

    def _get_instance(self, **kwargs):
        return ImplicitSpanMention(**kwargs)

    # We redefine these to use default semantics, overriding the operators
    # inherited from TemporarySpan
    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return id(self)



import numpy as np
import torch
import torch.nn as nn

from fonduer.learning.disc_learning import NoiseAwareModel
from fonduer.learning.disc_models.layers.rnn import RNN
from fonduer.learning.disc_models.layers.sparse_linear import SparseLinear
from fonduer.learning.disc_models.utils import (
    SymbolTable,
    mark_sentence,
    mention_to_tokens,
    pad_batch,
)
from fonduer.utils.config import get_config


class SparseLSTM(NoiseAwareModel):
    """
    LSTM model.

    :param name: User-defined name of the model
    :type name: str
    """

    def forward(self, x, x_idx, f, w):
        """Forward function.

        :param x: The sequence input (batch) of the model.
        :type x: list of torch.Tensor of shape (sequence_len * batch_size)
        :param x_idx: The feature indices for lstm output (batch) of the model.
        :type x_idx: torch.Tensor of shape (lstm_out_size * batch_size)
        :param f: The feature input of the model.
        :type f: torch.Tensor of shape (batch_size * feature_size)
        :param w: The input feature weight (batch) of the model.
        :type w: torch.Tensor of shape (batch_size, num_classes)
        :return: The output of LSTM layer.
        :rtype: torch.Tensor of shape (batch_size, num_classes)
        """

        batch_size = len(f)

        outputs = (
            torch.Tensor([]).cuda()
            if self.settings["host_device"] in self._gpu
            else torch.Tensor([])
        )

        # Calculate textual features from LSTMs
        for i in range(len(x)):
            state_word = self.lstms[0].init_hidden(batch_size)
            output = self.lstms[0].forward(x[i][0], x[i][1], state_word)
            outputs = torch.cat((outputs, output), 1)

        # Concatenate textual features with multi-modal features
        feaures = torch.cat((x_idx, f), 1)
        weights = torch.cat((outputs, w), 1)

        return self.sparse_linear(feaures, weights)

    def _check_input(self, X):
        """Check input format.

        :param X: The input data of the model.
        :type X: (candidates, features) pair
        :return: True if valid, otherwise False.
        :rtype: bool
        """
        return isinstance(X, tuple)

    def _preprocess_data(self, X, Y=None, idxs=None, train=False):
        """
        Preprocess the data:
        1. Make sentence with mention into sequence data for LSTM.
        2. Select subset of the input if idxs exists.

        :param X: The input data of the model.
        :type X: pair with candidates and corresponding features
        :param Y: The labels of input data (optional).
        :type Y: list of floats if num_classes = 2
            otherwise num_classes-length numpy array
        :param idxs: The selected indexs of input data.
        :type idxs: list or numpy.array
        :param train: An indicator for word dictionary to extend new words.
        :type train: bool
        :return: Preprocessed data.
        :rtype: list of (candidate, features) pairs
        """

        C, F = X

        # Create word dictionary for LSTM
        if not hasattr(self, "word_dict"):
            self.word_dict = SymbolTable()
            arity = len(C[0])
            # Add paddings into word dictionary
            for i in range(arity):
                # TODO: optimize this
                list(map(self.word_dict.get, ["~~[[" + str(i), str(i) + "]]~~"]))

        # Make sequence input for LSTM from candidates
        seq_data = []
        for candidate in C:
            cand_idx = []
            for i in range(len(candidate)):
                # Add mark for each mention in the original sentence
                args = [
                    (
                        candidate[i].context.get_word_start_index(),
                        candidate[i].context.get_word_end_index(),
                        i,
                    )
                ]
                s = mark_sentence(mention_to_tokens(candidate[i]), args)
                f = self.word_dict.get if train else self.word_dict.lookup
                cand_idx.append(list(map(f, s)))
            seq_data.append(cand_idx)

        # Generate proprcessed the input
        if idxs is None:
            if Y is not None:
                return (
                    [
                        (
                            seq_data[i],
                            F.indices[F.indptr[i] : F.indptr[i + 1]],
                            F.data[F.indptr[i] : F.indptr[i + 1]],
                        )
                        for i in range(len(C))
                    ],
                    Y,
                )
            else:
                return [
                    (
                        seq_data[i],
                        F.indices[F.indptr[i] : F.indptr[i + 1]],
                        F.data[F.indptr[i] : F.indptr[i + 1]],
                    )
                    for i in range(len(C))
                ]
        if Y is not None:
            return (
                [
                    (
                        seq_data[i],
                        F.indices[F.indptr[i] : F.indptr[i + 1]],
                        F.data[F.indptr[i] : F.indptr[i + 1]],
                    )
                    for i in idxs
                ],
                Y[idxs],
            )
        else:
            return [
                (
                    seq_data[i],
                    F.indices[F.indptr[i] : F.indptr[i + 1]],
                    F.data[F.indptr[i] : F.indptr[i + 1]],
                )
                for i in idxs
            ]

    def _update_settings(self, X):
        """
        Update the model argument.

        :param X: The input data of the model.
        :type X: list of (candidate, features) pairs
        """

        self.logger.info("Load defalut parameters for Sparse LSTM")
        config = get_config()["learning"]["SparseLSTM"]

        for key in config.keys():
            if key not in self.settings:
                self.settings[key] = config[key]

        self.settings["relation_arity"] = len(X[0][0])
        self.settings["lstm_dim"] = (
            len(X[0][0])
            * self.settings["hidden_dim"]
            * (2 if self.settings["bidirectional"] else 1)
        )

        # Add one feature for padding vector (all 0s)
        self.settings["input_dim"] = (
            X[1].shape[1]
            + len(X[0][0])
            * self.settings["hidden_dim"]
            * (2 if self.settings["bidirectional"] else 1)
            + 1
        )

    def _build_model(self):
        """
        Build the model.
        """
        # Set up LSTM modules
        self.lstms = nn.ModuleList(
            [
                RNN(
                    num_classes=0,
                    num_tokens=self.word_dict.s,
                    emb_size=self.settings["emb_dim"],
                    lstm_hidden=self.settings["hidden_dim"],
                    attention=self.settings["attention"],
                    dropout=self.settings["dropout"],
                    bidirectional=self.settings["bidirectional"],
                    use_cuda=self.settings["host_device"] in self._gpu,
                )
            ]
            * self.settings["relation_arity"]
        )

        if "input_dim" not in self.settings:
            raise ValueError("Model parameter input_dim cannot be None.")

        cardinality = self.cardinality if self.cardinality > 2 else 1

        # Set up final linear layer
        self.sparse_linear = SparseLinear(
            self.settings["input_dim"], cardinality, self.settings["bias"]
        )

    def _calc_logits(self, X, batch_size=None):
        """
        Calculate the logits.

        :param X: The input data of the model.
        :type X: list of (candidate, features) pairs
        :param batch_size: The batch size.
        :type batch_size: int
        :return: The output logits of model.
        :rtype: torch.Tensor of shape (batch_size, num_classes) if num_classes > 2
            otherwise shape (batch_size, 1)
        """

        # Generate LSTM input
        C = np.array(list(zip(*X))[0])

        # Check LSTM input dimension size matches the number of lstms in the model
        assert len(C[0]) == len(self.lstms)

        # Generate sparse multi-modal feature input
        F = (
            np.array(list(zip(*X))[1]) + self.settings["lstm_dim"] + 1
        )  # Correct the index since 0 is the padding and placeholder for lstm feature
        V = np.array(list(zip(*X))[2])

        outputs = (
            torch.Tensor([]).cuda()
            if self.settings["host_device"] in self._gpu
            else torch.Tensor([])
        )

        n = len(F)
        if batch_size is None:
            batch_size = n
        for batch_st in range(0, n, batch_size):
            batch_ed = batch_st + batch_size if batch_st + batch_size <= n else n

            # TODO: optimize this
            sequences = []
            # For loop each relation arity
            for i in range(len(C[0])):
                sequence = []
                # Generate sequence for the batch
                for j in range(batch_st, batch_ed):
                    sequence.append(C[j][i])
                x, x_mask = pad_batch(sequence, self.settings["max_sentence_length"])
                if self.settings["host_device"] in self._gpu:
                    x = x.cuda()
                    x_mask = x_mask.cuda()
                sequences.append((x, x_mask))

            lstm_weight_indices = torch.as_tensor(
                np.arange(1, self.settings["lstm_dim"] + 1)
            ).repeat(batch_ed - batch_st, 1)

            features, _ = pad_batch(F[batch_st:batch_ed], 0)
            values, _ = pad_batch(V[batch_st:batch_ed], 0, type="float")

            if self.settings["host_device"] in self._gpu:
                lstm_weight_indices = lstm_weight_indices.cuda()
                features = features.cuda()
                values = values.cuda()

            output = self.forward(sequences, lstm_weight_indices, features, values)
            if self.cardinality == 2:
                outputs = torch.cat((outputs, output.view(-1)), 0)
            else:
                outputs = torch.cat((outputs, output), 0)

        return outputs



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporarySectionMention(TemporaryContext):
    """The TemporaryContext version of SectionMention."""

    def __init__(self, section):
        super(TemporarySectionMention, self).__init__()
        self.section = section  # The section Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.section == other.section
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.section != other.section
        except AttributeError:
            return True

    def __contains__(self, other_section):
        return self.__eq__(other_section)

    def __hash__(self):
        return hash(self.section)

    def get_stable_id(self):
        """Return a stable id for the ``SectionMention``."""
        return "%s::%s:%s" % (
            self.section.document.name,
            self._get_polymorphic_identity(),
            self.section.position,
        )

    def _get_table(self):
        return SectionMention

    def _get_polymorphic_identity(self):
        return "section_mention"

    def _get_insert_args(self):
        return {"section_id": self.section.id}

    def __repr__(self):
        return "{}(document={}, position={})".format(
            self.__class__.__name__, self.section.document.name, self.section.position
        )

    def _get_instance(self, **kwargs):
        return TemporarySectionMention(**kwargs)


class SectionMention(Context, TemporarySectionMention):
    """A section ``Mention``."""

    __tablename__ = "section_mention"

    #: The unique id of the ``SectionMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Section``.
    section_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Section``.
    section = relationship("Context", foreign_keys=section_id)

    __table_args__ = (UniqueConstraint(section_id),)

    __mapper_args__ = {
        "polymorphic_identity": "section_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryFigureMention(TemporaryContext):
    """The TemporaryContext version of FigureMention."""

    def __init__(self, figure):
        super(TemporaryFigureMention, self).__init__()
        self.figure = figure  # The figure Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.figure == other.figure
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.figure != other.figure
        except AttributeError:
            return True

    def __contains__(self, other_span):
        return self.__eq__(other_span)

    def __hash__(self):
        return hash(self.figure)

    def get_stable_id(self):
        """Return a stable id for the ``FigureMention``."""
        return "%s::%s:%s" % (
            self.figure.document.name,
            self._get_polymorphic_identity(),
            self.figure.position,
        )

    def _get_table(self):
        return FigureMention

    def _get_polymorphic_identity(self):
        return "figure_mention"

    def _get_insert_args(self):
        return {"figure_id": self.figure.id}

    def __repr__(self):
        return "{}(document={}, position={}, url={})".format(
            self.__class__.__name__,
            self.figure.document.name,
            self.figure.position,
            self.figure.url,
        )

    def _get_instance(self, **kwargs):
        return TemporaryFigureMention(**kwargs)


class FigureMention(Context, TemporaryFigureMention):
    """A figure ``Mention``."""

    __tablename__ = "figure_mention"

    #: The unique id of the ``FigureMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Figure``.
    figure_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Figure``.
    figure = relationship("Context", foreign_keys=figure_id)

    __table_args__ = (UniqueConstraint(figure_id),)

    __mapper_args__ = {
        "polymorphic_identity": "figure_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryParagraphMention(TemporaryContext):
    """The TemporaryContext version of ParagraphMention."""

    def __init__(self, paragraph):
        super(TemporaryParagraphMention, self).__init__()
        self.paragraph = paragraph  # The paragraph Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.paragraph == other.paragraph
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.paragraph != other.paragraph
        except AttributeError:
            return True

    def __contains__(self, other_paragraph):
        return self.__eq__(other_paragraph)

    def __hash__(self):
        return hash(self.paragraph)

    def get_stable_id(self):
        """Return a stable id for the ``ParagraphMention``."""
        return "%s::%s:%s" % (
            self.paragraph.document.name,
            self._get_polymorphic_identity(),
            self.paragraph.position,
        )

    def _get_table(self):
        return ParagraphMention

    def _get_polymorphic_identity(self):
        return "paragraph_mention"

    def _get_insert_args(self):
        return {"paragraph_id": self.paragraph.id}

    def __repr__(self):
        return "{}(document={}, position={})".format(
            self.__class__.__name__,
            self.paragraph.document.name,
            self.paragraph.position,
        )

    def _get_instance(self, **kwargs):
        return TemporaryParagraphMention(**kwargs)


class ParagraphMention(Context, TemporaryParagraphMention):
    """A paragraph ``Mention``."""

    __tablename__ = "paragraph_mention"

    #: The unique id of the ``ParagraphMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Paragraph``.
    paragraph_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Paragraph``.
    paragraph = relationship("Context", foreign_keys=paragraph_id)

    __table_args__ = (UniqueConstraint(paragraph_id),)

    __mapper_args__ = {
        "polymorphic_identity": "paragraph_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryDocumentMention(TemporaryContext):
    """The TemporaryContext version of DocumentMention."""

    def __init__(self, document):
        super(TemporaryDocumentMention, self).__init__()
        self.document = document  # The document Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.document == other.document
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.document != other.document
        except AttributeError:
            return True

    def __contains__(self, other_document):
        return self.__eq__(other_document)

    def __hash__(self):
        return hash(self.document)

    def get_stable_id(self):
        """Return a stable id for the ``DocumentMention``."""
        return "%s::%s" % (self.document.name, self._get_polymorphic_identity())

    def _get_table(self):
        return DocumentMention

    def _get_polymorphic_identity(self):
        return "document_mention"

    def _get_insert_args(self):
        return {"document_id": self.document.id}

    def __repr__(self):
        return "{}(document={})".format(self.__class__.__name__, self.document.name)

    def _get_instance(self, **kwargs):
        return TemporaryDocumentMention(**kwargs)


class DocumentMention(Context, TemporaryDocumentMention):
    """A document ``Mention``."""

    __tablename__ = "document_mention"

    #: The unique id of the ``DocumentMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Document``.
    document_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Document``.
    document = relationship("Context", foreign_keys=document_id)

    __table_args__ = (UniqueConstraint(document_id),)

    __mapper_args__ = {
        "polymorphic_identity": "document_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import relationship

from fonduer.candidates.models.temporary_context import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryCaptionMention(TemporaryContext):
    """The TemporaryContext version of CaptionMention."""

    def __init__(self, caption):
        super(TemporaryCaptionMention, self).__init__()
        self.caption = caption  # The caption Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return self.caption == other.caption
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.caption != other.caption
        except AttributeError:
            return True

    def __contains__(self, other_caption):
        return self.__eq__(other_caption)

    def __hash__(self):
        return hash(self.caption)

    def get_stable_id(self):
        """Return a stable id for the ``CaptionMention``."""
        return "%s::%s:%s" % (
            self.caption.document.name,
            self._get_polymorphic_identity(),
            self.caption.position,
        )

    def _get_table(self):
        return CaptionMention

    def _get_polymorphic_identity(self):
        return "caption_mention"

    def _get_insert_args(self):
        return {"caption_id": self.caption.id}

    def __repr__(self):
        return "{}(document={}, position={})".format(
            self.__class__.__name__, self.caption.document.name, self.caption.position
        )

    def _get_instance(self, **kwargs):
        return TemporaryCaptionMention(**kwargs)


class CaptionMention(Context, TemporaryCaptionMention):
    """A caption ``Mention``."""

    __tablename__ = "caption_mention"

    #: The unique id of the ``CaptionMention``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Caption``.
    caption_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))

    #: The parent ``Caption``.
    caption = relationship("Context", foreign_keys=caption_id)

    __table_args__ = (UniqueConstraint(caption_id),)

    __mapper_args__ = {
        "polymorphic_identity": "caption_mention",
        "inherit_condition": (id == Context.id),
    }

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()
